import time
from typing import List, Union, Optional
from datetime import datetime, timedelta

from redis import Redis
from pydantic import ValidationError

from blocks.types import Event, Source
from blocks.logger import logger
from blocks.redis.serdes import Deserializer, deserialize
from blocks.redis.streams import InputStream


class RedisConsumer(Source):
    """
    Class represents event source that reads messages from Redis streams
    and wraps them into given events. On initialization must get list
    of input streams in order to perform arbitrary mapping. In most cases
    you don't need to use RedisConsumer directly, RedisApp can make this for you.

    Example::

      >>> from typing import NamedTuple

      >>> from redis import Redis
      >>> from blocks import Event, Graph
      >>> from blocks.redis import RedisConsumer, InputStream
      >>>
      >>> class MyEvent(NamedTuple):
      ...     x: int
      >>>
      >>> streams = [InputStream('some_stream', MyEvent)]
      >>> client = Redis(...)
      >>> blocks = (RedisConsumer(client, streams, read_timeout=100), ...)
    """

    def __init__(
        self,
        client: Redis,
        streams: List[InputStream],
        read_timeout: int,
        deserializer: Deserializer = deserialize,
    ) -> None:
        self._client = client
        self._codecs = {stream.name: stream.event for stream in streams}
        self._offsets = {stream.name.encode(): stream.start_id.encode() for stream in streams}
        self._count = min(stream.messages_limit for stream in streams)
        self._last_call: Optional[datetime] = None
        self._read_timeout = read_timeout
        self._deserializer = deserializer

        out_annots = {stream.event for stream in streams}
        self.__call__.__annotations__['return'] = List[Union[tuple(out_annots)]]  # type: ignore

    # ToDo (tribunsky.kir): complexity? Author didn't even care about complexity.
    def __call__(self) -> List[Event]:
        events = []
        if self._last_call is None or (datetime.now() - self._last_call) > timedelta(milliseconds=self._read_timeout):
            try:
                all_data = self._client.xread(self._offsets, self._count)  # type: ignore
                self._last_call = datetime.now()
            except Exception as exc:
                logger.error(exc)
            else:
                for stream_name, msgs in all_data:
                    for offset, message in msgs:
                        deserialized = self._deserializer(message)
                        codec = self._codecs[stream_name.decode()]
                        try:
                            events.append(codec(**deserialized))
                        except ValidationError as err:
                            logger.warning(err)
                            logger.warning('Deserialized message: {0}'.format(deserialized))
                        self._offsets[stream_name] = offset
        else:
            to_sleep = timedelta(milliseconds=self._read_timeout) - (datetime.now() - self._last_call)
            time.sleep(to_sleep.microseconds / 1000000)
        return events

    def close(self) -> None:
        """Graceful shutdown: close reids client."""
        self._client.close()



import time
from typing import Type

from blocks import Event


class _Stream:
    def __init__(self, name: str, event: Type[Event]) -> None:
        self.name = name
        self.event = event

# ToDo (Daniil): Add NoNewEvents generation in consumer
# ToDo (Daniil): Add possibility to store offsets in Redis KV instead of app memory


class InputStream(_Stream):
    def __init__(
        self,
        name: str,
        event: Type[Event],
        *,
        start_id: str = f'{int(time.time() * 1000)}-0',
        messages_limit: int = 1000,
    ) -> None:
        super().__init__(name, event)
        self.messages_limit = messages_limit
        self.start_id = start_id


class OutputStream(_Stream):
    def __init__(self, name: str, event: Type[Event], max_len: int = 1000) -> None:
        super().__init__(name, event)
        self.max_len = max_len



"""This module contains some anti-boilerplate stuff for a typical Kafka-Kafka stream application."""

from typing import Type, Union, Optional, Sequence

from wunderkafka import ConsumerConfig, ProducerConfig

from blocks import App, Block, Event
from blocks.logger import logger
from blocks.validation import validate_blocks
from blocks.kafka.topics import InputTopic, OutputTopic
from blocks.kafka.sources import KafkaSource
from blocks.kafka.processors import CommitEvent, KafkaProducer, OffsetCommitter

Topic = Union[InputTopic, OutputTopic]


class KafkaApp(App):
    """
    High level API for building applications with Kafka topics as inputs and outputs.

    Also adds handlers for common events for kafka-specific things like commiting offsets.

    Example::

      >>> # Very dummy replicator
      >>> from pydantic import BaseModel
      >>> from wunderkafka import ConsumerConfig, ProducerConfig

      >>> from blocks import Event, processor
      >>> from blocks.kafka import KafkaApp, InputTopic, OutputTopic

      >>> class MyEvent(BaseModel):
      ...     x: int

      >>> @processor
      ... def printer(e: MyEvent) -> None:
      ...     print(e)

      >>> consumer_config = ConsumerConfig(...)
      >>> producer_config = ProducerConfig(...)
      >>> topics = [
      ...     InputTopic('my_topic', MyEvent, from_beginning=False),
      ...     OutputTopic('my_other_topic', MyEvent),
      ... ]
      >>> blocks = (printer(),)
      >>> KafkaApp(topics, blocks, consumer_config, producer_config).run()
    """

    def __init__(
        self,
        topics: Sequence[Topic],
        blocks: Sequence[Block],
        consumer_config: Optional[ConsumerConfig] = None,
        producer_config: Optional[ProducerConfig] = None,
        terminal_event: Optional[Type[Event]] = None,
    ) -> None:
        """
        Initiate Kafka app instance.

        :param topics:              Input/OutputTopic to work with.
        :param blocks:              Additional blocks with business logic, other events sources or just processors
                                    to make effects.
        :param consumer_config:     Default consumer config to be applied, if not defined per InputTopic.
        :param producer_config:     Default producer config to be applied, if not defined per OutputTopic.
        :param terminal_event:      If specified, special event to notify app to stop receiving events and close all
                                    blocks properly.
        """
        super().__init__(blocks=[], terminal_event=terminal_event)
        for block in blocks:
            self._graph.add_block(block)

        input_topics = []
        output_topics = []
        for topic in topics:
            # ToDo (tribunsky.kir): check that there is no multi-inherited InputOutputTopic?
            if isinstance(topic, InputTopic):
                input_topics.append(topic)
            elif isinstance(topic, OutputTopic):
                output_topics.append(topic)
            else:
                logger.warning('{0} is not a valid InputTopic/OutputTopic, ignoring it...'.format(topic))

        if output_topics:
            self._graph.add_block(KafkaProducer(output_topics, producer_config))

        if input_topics:
            source = KafkaSource(input_topics, consumer_config)
            self._graph.add_block(source)
            if CommitEvent in self._graph.outputs:
                self._graph.add_block(OffsetCommitter(source.consumers))
        else:
            # ToDo (tribunsky.kir): or should we raise Value Error cause it's KafkaApp, not any hybrid graph.
            #                       or we may check other blocks here if there are any sources.
            logger.warning('No input topics detected. Please, make sure, that you have other sources in your code.')

        # ToDo (tribunsky.kir): "Great" API! Move validation back to the graph
        validate_blocks(self._graph.blocks)



import os.path
import sqlite3
from dataclasses import dataclass

import pytest

from blocks.db.types import Row
from blocks.db.next.sql import Query, Dialects
from blocks.sqlite import SQLiteWriter, SQLiteReader


@dataclass
class TableRow(Row):
    id: int
    name: str
    text: str


@pytest.fixture
def connect():
    def connection_factory():
        con = sqlite3.connect('example.db')
        with con:
            con.execute('''CREATE TABLE IF NOT EXISTS test_table (id, name, text)''')
        return con

    return connection_factory


# FixMe (tribunski.kir): potentially dangerous, use smth like mkstemp
@pytest.fixture(autouse=True)
def run_around_tests():
    if os.path.exists('example.db'):
        raise
    yield
    os.remove('example.db')


def test_read_write(connect) -> None:
    read_query = Query(TableRow, dialect=Dialects.sqlite).SELECT('*').FROM('test_table')
    reader = SQLiteReader([read_query], connect)
    events = reader()
    assert events == []

    insert_query = Query(dialect=Dialects.sqlite).INSERT(TableRow).INTO('test_table')
    writer = SQLiteWriter([insert_query], connect)
    row_original = TableRow(id=1, name='one', text='test')
    writer(row_original)

    [event] = reader()
    assert event == row_original

    update_query = Query(dialect=Dialects.sqlite).UPDATE('test_table').SET(TableRow).WHERE('id')
    writer = SQLiteWriter([update_query], connect)
    row_updated = TableRow(id=1, name='one', text='updated')

    writer(row_updated)

    [event] = reader()
    assert event == row_updated



"""Handful decorators to create *Sources* and *Processors* quickly."""

from typing import Any, Type, Callable, Optional, Awaitable

from blocks.types import Event, Source, Processor, AsyncSource, EventOrEvents, AsyncProcessor, ParallelEvent

# ToDo (tribunsky.kir): redo annotations to avoid "error: Untyped decorator makes function <...>" untyped
# ToDo (tribunsky.kir): remove closures and explicit code duplication

SourceFunction = Callable[[], EventOrEvents]
SourceAwaitable = Callable[[], Awaitable[EventOrEvents]]
ProcessorFunction = Callable[[Any], Optional[EventOrEvents]]
ProcessorAwaitable = Callable[[Any], Awaitable[Optional[EventOrEvents]]]


def source(function: SourceFunction) -> Type[Source]:
    """
    Make a Source from the decorated function.

    Example::

      >>> from dataclasses import dataclass

      >>> from blocks import source


      >>> @dataclass
      ... class E:
      ...     x: int


      >>> @source
      >>> def generator() -> E:
      ...     return E(1)


      >>> blocks = (generator(), ...)

    :param function:    Given function with no arguments.
                        Function should return event or sequence of events to have an effect.
    :return:            Factory which creates a Source.
    """

    def _call(self: Source) -> EventOrEvents:
        return function()

    T = type(f'{function.__name__}', (Source, ), {'__call__': _call})
    T.__call__.__annotations__ = function.__annotations__
    return T


def processor(function: ProcessorFunction) -> Type[Processor]:
    """
    Make a Processor from the decorated function.

    Example::

      >>> from typing import Event

      >>> from blocks import processor

      >>> class MyEvent(Event):
      ...     ...

      >>> @processor
      ... def printer(event: MyEvent) -> None:
      ...     print(event)

      >>> blocks = (printer(), ...)

    :param function:    Given function with a single argument (event).
                        Function may return None, event or sequence of events.
    :return:            Factory which creates a Processor.
    """

    def _call(self: Processor, event: Event) -> Optional[EventOrEvents]:
        return function(event)

    T = type(f'{function.__name__}', (Processor, ), {'__call__': _call})
    T.__call__.__annotations__ = function.__annotations__
    return T


def parallel_processor(function: ProcessorFunction,
                       *,
                       timeout: float = 5.0,
                       daemon: bool = True,
                       force_terminating: bool = True
                       ) -> Type[Processor]:
    """
    Make a Parallel event from the decorated function.

    Example::
      >>> from dataclasses import dataclass

      >>> from blocks import processor

      >>> @dataclass
      >>> class MyEvent:
      ...     ...

      >>> @parallel_processor
      ... def printer(event: MyEvent) -> None:
      ...     print(event)

      >>> blocks = (printer(), ...)

    :param function:            Given function with a single argument (event).
                                Function may return None, event or sequence of events.

    param timeout:              Param for join process.
    param daemon:               Allows us daemon processes.
    param force_terminating:    Allows force termiante process if this didn`t end by timeout
    :return:                    Factory which creates a Processor.
    """
    def _call(self: Processor, event: Event) -> Optional[EventOrEvents]:
        return ParallelEvent(
            function=function,
            trigger=event,
            timeout=timeout,
            daemon=daemon,
            force_terminating=force_terminating
        )

    T = type(f'{function.__name__}', (Processor, ), {'__call__': _call})
    T.__call__.__annotations__ = function.__annotations__
    return T


def async_source(function: SourceAwaitable) -> Type[AsyncSource]:
    """
    Make an awaitable Source from the decorated awaitable function.

    :param function:    Given function with no arguments.
                        Function should return event or sequence of events to have an effect.
    :return:            Factory which creates an awaitable Source.
    """

    async def _call(self: AsyncSource) -> EventOrEvents:
        return await function()

    T = type(f'{function.__name__}', (AsyncSource, ), {'__call__': _call})
    T.__call__.__annotations__ = function.__annotations__
    return T


def async_processor(function: ProcessorAwaitable) -> Type[AsyncProcessor]:
    """
    Make an awaitable Processor from the decorated awaitable function.

    :param function:    Given function with a single argument (event).
                        Function may return None, event or sequence of events.
    :return:            Factory which creates an awaitable Processor.
    """

    async def _call(self: AsyncProcessor, event: Event) -> Optional[EventOrEvents]:
        return await function(event)

    T = type(f'{function.__name__}', (AsyncProcessor, ), {'__call__': _call})
    T.__call__.__annotations__ = function.__annotations__
    return T



# Hardly reusing & dirty hacking this https://death.andgravity.com/query-builder-how

from enum import Enum
from types import MappingProxyType
from typing import Any, Optional, Type, Union

# ToDo (tribunsky.kir): allow different base classes for relational DB
from dataclasses import fields

import functools
import textwrap

from blocks.db.types import Row, Table
from blocks.logger import logger


class Dialects(str, Enum):
    postgres: str = 'postgres'
    sqlite: str = 'sqlite'


_PLACEHOLDERS = MappingProxyType({
    Dialects.postgres: '%s',
    Dialects.sqlite: '?',
})

# Ordering matters =/
KEYWORDS = (
    'INSERT',
    'INTO',
    'UPDATE',
    'SELECT',
    'FROM',
    'SET',
    'WHERE',
    'ORDER BY',
    'LIMIT',
)

DEFAULT_SEPARATOR = ','


class Query(object):

    def __init__(self, codec: Optional[Type[Union[Row, Table]]] = None, dialect: str = Dialects.postgres) -> None:
        message = [
            'You are using EXPERIMENTAL API for queries',
            'which is currently unstable and absolutely buggy & unsafe!',
            'Use it on your own risk!'
        ]
        logger.warning(' '.join(message))
        self.data = {k: [] for k in KEYWORDS}
        self._whereable = set()
        self.codec: Type[Union[Row, Table]] = codec
        self._fields = []
        self._placeholder = _PLACEHOLDERS[dialect]

    @property
    def text(self) -> str:
        return str(self)

    def parametrize(self, event: Union[Row, Table]) -> tuple:
        return tuple([getattr(event, f) for f in self._fields])

    def add(self, keyword: str, *args) -> 'Query':
        target = self.data[keyword]

        # ToDo (tribunsky.kir): Separate different verbs or enter some simple grammar
        # ToDo (tribunsky.kir): Do not put placeholders as strings while adding parts of query,
        #                       use some stub and
        for arg in args:
            if keyword == 'INTO' and self.data.get('INSERT'):
                insert = self.data.get('INSERT')
                if insert:
                    statement = ['INTO'] + [arg] + [', '.join(insert)]
                    self.data['INSERT'] = [' '.join(statement)]
                else:
                    target.append(_clean_up(arg))
            elif isinstance(arg, str):
                if keyword == 'WHERE' and arg in self._whereable:
                    fld = _clean_up(arg)
                    target.append(fld + (' = ' + self._placeholder))
                    self._fields.append(fld)
                else:
                    target.append(_clean_up(arg))
            else:
                self.codec = arg
                self._whereable = set(field.name for field in fields(arg))
                if keyword == 'INSERT':
                    attrs = [attr.name for attr in fields(arg)]
                    flds = []
                    vals = []
                    for ix, attr in enumerate(attrs, 1):
                        self._fields.append(attr)
                        if ix == 1:
                            flds.append('(' + attr)
                            vals.append(('(' + self._placeholder))
                        elif ix == len(attrs):
                            flds.append(attr + ')')
                            vals.append((self._placeholder + ')'))
                        else:
                            flds.append(attr)
                            vals.append((self._placeholder))

                    into_table = self.data.get('INTO')
                    if into_table:
                        self.data['INSERT'] = ['INTO'] + into_table + flds
                    else:
                        self.data['INSERT'] = flds
                    self.data['VALUES'] = [', '.join(vals)]
                else:
                    if keyword == 'SET':
                        to_append = []
                        for field in fields(arg):
                            to_append.append(field.name + (' = ' + self._placeholder))
                            self._fields.append(field.name)
                        target += to_append
                    else:
                        target += [field.name for field in fields(arg)]

        return self

    def __getattr__(self, name: str) -> Any:
        if not name.isupper():
            return getattr(super(), name)
        return functools.partial(self.add, name.replace('_', ' '))

    def __str__(self) -> str:
        return ''.join(self._lines())

    def _lines(self):
        for keyword, things in self.data.items():
            if not things:
                continue

            yield '{0}\n'.format(keyword)
            yield from self._lines_keyword(keyword, things)

    def _lines_keyword(self, keyword: str, things):
        for i, thing in enumerate(things, 1):
            last = i == len(things)

            yield self._indent(thing)

            if not last:
                if keyword == 'WHERE':
                    yield ' AND '
                else:
                    yield DEFAULT_SEPARATOR

            yield '\n'

    _indent = functools.partial(textwrap.indent, prefix='    ')


def _clean_up(thing: str) -> str:
    return textwrap.dedent(thing.rstrip()).strip()



from dataclasses import dataclass
from typing import Callable, Any, Optional

from blocks.types.base import Event, EventOrEvents


@dataclass
class ParallelEvent(Event):
    function: Callable[[Any], Optional[EventOrEvents]]
    trigger: EventOrEvents
    timeout: float
    daemon: bool
    force_terminating: bool



from typing import Any, Dict, List, Type

from blocks.types import Block, Event


class AnnotationError(Exception):
    ...


def _get_annotations(block: Block) -> Dict[str, Any]:
    return dict(block.__call__.__annotations__)


def get_input_events_type(block: Block) -> List[Type[Event]]:
    annots = _get_annotations(block)
    annots.pop('return', None)
    if len(annots) == 0:
        return []
    input_type = list(annots.values())[0]
    generic_args = getattr(input_type, '__args__', None)
    if generic_args is not None:
        return list(generic_args)
    return [input_type]


def get_output_events_type(block: Block) -> List[Type[Event]]:
    annots = _get_annotations(block)
    annotations_to_flatten = [annots.get('return', None)]
    result = []
    while annotations_to_flatten:
        annotation = annotations_to_flatten.pop(0)
        if annotation is not None:
            generic_args = getattr(annotation, '__args__', None)
            if generic_args is not None:
                annotations_to_flatten.extend(list(generic_args))
            else:
                result.append(annotation)
    return result



from typing import Any, Dict
from dataclasses import asdict, is_dataclass

from blocks.types import Event

try:
    from pydantic import BaseModel
except ImportError:
    HAS_PYDANTIC = False
else:
    HAS_PYDANTIC = True


def event2dict(event: Event) -> Dict[str, Any]:
    if HAS_PYDANTIC is True:
        if isinstance(event, BaseModel):
            return event.dict()
    else:
        method = getattr(event, 'dict')
        if callable(method):
            return event.dict()  # type: ignore
    if is_dataclass(event):
        return asdict(event)
    if isinstance(event, Event):
        return {attr: getattr(event, attr) for attr in dir(event) if not attr.startswith('__')}



from typing import Dict, List, Type, Union

from redis import Redis

from blocks import Event, Processor
from blocks.utils import event2dict
from blocks.redis.serdes import Serializer, serialize
from blocks.redis.streams import OutputStream


class RedisProducer(Processor):
    """
    Class represents event processor that writes messages to Redis streams
    based on data received from event. On initialization must get list
    of output streams in order to perform arbitrary mapping. In most cases
    you don't need to use RedisProducer directly, RedisApp can make this for you.

    Example::

      >>> from typing import NamedTuple

      >>> from redis import Redis
      >>> from blocks import Event, Graph
      >>> from blocks.redis import RedisProducer, OutputStream

      >>> class MyEvent(NamedTuple):
      ...     x: int

      >>> streams = [OutputStream('some_stream', MyEvent)]
      >>> client = Redis(...)
      >>> blocks = (RedisProducer(client, streams), ...)
    """

    def __init__(
        self,
        client: Redis,
        streams: List[OutputStream],
        serializer: Serializer = serialize
    ) -> None:
        self._client = client
        self._serializer = serializer

        self._streams: Dict[Type[Event], OutputStream] = {stream.event: stream for stream in streams}
        self.__call__.__annotations__.update(
            {
                'event': Union[tuple(stream.event for stream in streams)],
            },
        )

    def __call__(self, event: Event) -> None:
        stream = self._streams[type(event)]
        serialized = self._serializer(event2dict(event))
        self._client.xadd(stream.name, fields=serialized, maxlen=stream.max_len)  # type: ignore

    def close(self) -> None:
        """Graceful shutdown: close reids client."""
        self._client.close()



import datetime
from textwrap import dedent
from dataclasses import dataclass

import pytest

from blocks.db.next.sql import Query, Dialects
from blocks.db import Row


@dataclass
class TableRow(Row):
    id: int
    name: str
    dt: datetime.datetime


@pytest.fixture
def row() -> TableRow:
    return TableRow(id=1, name='test', dt=datetime.datetime.now())


# ToDo (tribunsky.kir): looks like there is a nice to be a query's method/representation
def check(query: Query, answer: str) -> None:
    q = str(query)
    delim = ' '
    normalized = dedent(delim.join([el.strip() for el in q.replace('\n', delim).split(delim) if el.strip()])).casefold()
    assert normalized == answer.casefold()


def test_select_all() -> None:
    query = Query(dialect=Dialects.sqlite).SELECT('*').FROM('table')
    answer = 'SELECT * FROM table'
    check(query, answer)


def test_select_model() -> None:
    query = Query(dialect=Dialects.sqlite).SELECT(TableRow).FROM('table')
    answer = 'SELECT id, name, dt from table'
    check(query, answer)


def test_select_model_where() -> None:
    query = Query(dialect=Dialects.sqlite).SELECT(TableRow).FROM('table').WHERE('name is not null', 'dt is null')
    answer = 'SELECT id, name, dt from table where name is not null and dt is null'
    check(query, answer)


def test_insert(row) -> None:
    query = Query(dialect=Dialects.sqlite).INSERT(TableRow).INTO('table')
    answer = 'INSERT INTO table (id, name, dt) VALUES (?, ?, ?)'
    check(query, answer)

    tpl = query.parametrize(row)
    assert tpl == (row.id, row.name, row.dt)


def test_update_where_self_reference(row) -> None:
    query = Query(dialect=Dialects.sqlite).UPDATE('table').SET(TableRow).WHERE('id')
    answer = 'UPDATE table SET id = ?, name = ?, dt = ? WHERE id = ?'
    check(query, answer)

    tpl = query.parametrize(row)
    assert tpl == (row.id, row.name, row.dt, row.id)


def test_update_where_self_reference_multiple(row):
    query = Query(dialect=Dialects.sqlite).UPDATE('table').SET(TableRow).WHERE('id', 'name is not null')
    answer = 'UPDATE table SET id = ?, name = ?, dt = ? WHERE id = ? AND name is not null'
    check(query, answer)

    tpl = query.parametrize(row)
    assert tpl == (row.id, row.name, row.dt, row.id)


def test_update_where_clause(row):
    query = Query(dialect=Dialects.sqlite).UPDATE('table').SET(TableRow).WHERE('id > 10')
    answer = 'UPDATE table SET id = ?, name = ?, dt = ? WHERE id > 10'
    check(query, answer)

    tpl = query.parametrize(row)
    assert tpl == (row.id, row.name, row.dt)



from blocks.app import App
from blocks.decorators import source, processor, async_source, async_processor, parallel_processor
from blocks.graph import Graph
from blocks.runners import Runner, AsyncRunner
from blocks.types import Block, Event, Source, Processor, AsyncSource, AsyncProcessor



import re
import subprocess
from collections import defaultdict
from typing import List, Union

from blocks.logger import logger
from blocks.types.graph import RenderingKernelType

try:
    from graphviz import ExecutableNotFound
    from graphviz import Digraph
except ImportError:
    HAS_GRAPHVIZ = False
else:
    HAS_GRAPHVIZ = True

try:
    import networkx as nx
except ImportError:
    HAS_NETWORKX = False
else:
    HAS_NETWORKX = True

try:
    from matplotlib import pyplot as plt
    from matplotlib.axes import Axes
except ImportError:
    HAS_MATPLOTLIB = False
else:
    HAS_MATPLOTLIB = True

from blocks.types.base import Block, AnyProcessors
from blocks.annotations import get_output_events_type


def build_graph(
    blocks: List[Block],
    processors_dict: AnyProcessors,
    rendering_type: Union[str, RenderingKernelType]
) -> None:
    if rendering_type == RenderingKernelType.matplotlib:
        _build_graph_matplot(blocks, processors_dict)
    else:
        _build_graph_graphviz(blocks, processors_dict)


def _build_graph_matplot(
    blocks: List[Block],
    processors_dict: AnyProcessors,
    *,
    file_name: str = 'GRAPH'
) -> None:
    if not HAS_MATPLOTLIB or not HAS_NETWORKX:
        logger.error('Install the NetworkX and matplotlib packages before using')
        raise RuntimeError('Install the NetworkX and matplotlib packages before using')

    event_to_processor = defaultdict(list)
    g = nx.DiGraph()
    for s in blocks:
        sn = type(s).__name__
        g.add_node(sn)
        out_events = get_output_events_type(s)
        for out_event in out_events:
            event_to_processor[out_event].append(sn)
    edge_labels = {}
    for event, processors in processors_dict.items():
        for processor in processors:
            processor_name = type(processor).__name__
            if event in event_to_processor:
                for proc in event_to_processor[event]:
                    g.add_edge(
                        proc,
                        processor_name,
                        length=400
                    )
                    edge_labels[(proc, processor_name)] = event.__name__
    pos = nx.spring_layout(g)
    ax: Axes = plt.subplot()
    ax.margins(0.2)

    nx.draw_networkx(
        g,
        pos,
        ax=ax,
        edge_color='black',
        node_color='pink',
        with_labels=True,
        arrows=True,
        font_size=8,
    )
    nx.draw_networkx_edge_labels(
        g, pos,
        font_color='red',
        edge_labels=edge_labels,
        font_size=8,
    )
    plt.axis('off')
    plt.savefig(file_name, dpi=100)


def _build_graph_graphviz(blocks: List[Block], processors_dict: AnyProcessors) -> None:
    if not HAS_GRAPHVIZ:
        logger.error('Install the graphviz package before using')
        raise RuntimeError('Install the graphviz package before using')
    output = subprocess.run(['dot', '-V'], capture_output=True, shell=True)
    result = re.search(r'graphviz version \d+?.\d+?.\d+?', output.stderr.decode('utf-8'))
    if result is None:
        raise Exception("Can't render graph visualization, "
                        "make sure that graphviz C-library installed in your system"
                        )
    logger.info('Using system {}'.format(result.group()))

    dag = Digraph(
        'DAG',
        format='svg',
        filename='DAG',
        node_attr={'color': 'lightblue2', 'style': 'filled'},
    )
    event_to_processor = defaultdict(list)
    for processor in blocks:
        processor_name = type(processor).__name__
        dag.node(processor_name)
        out_events = get_output_events_type(processor)
        for out_event in out_events:
            event_to_processor[out_event].append(processor_name)
    for event, processors in processors_dict.items():
        for processor in processors:
            processor_name = type(processor).__name__
            if event in event_to_processor:
                for proc in event_to_processor[event]:
                    dag.edge(
                        proc,
                        processor_name,
                        event.__name__,
                    )
    dag.render(cleanup=True)



# ToDo (tribunski.kir): change it to the protocol
from sqlite3 import Connection
from typing import List, Union, Callable

from blocks import Processor
from blocks.db.types import Row, Table
from blocks.db.next.sql import Query
from blocks.logger import logger


def _exec_queries(conn: Connection, query: Query, table: Table) -> None:
    query_formatted = str(query)
    insert_tuples = [query.parametrize(row) for row in table.rows]
    total_count = len(insert_tuples)

    cursor = conn.cursor()

    for ix, tpl in enumerate(insert_tuples, 1):
        logger.info('[{0}/{1}] Executing query: {2} {3}'.format(ix, total_count, query_formatted, tpl))
        cursor.execute(query_formatted, tpl)

    conn.commit()
    conn.close()


class SQLWriter(Processor):

    def __init__(self, queries: List[Query], connection_factory: Callable[[], Connection]) -> None:
        self._connection_factory = connection_factory
        self._queries = {query.codec: query for query in queries}
        self._closed = False
        self.__call__.__annotations__.update(
            {
                'event': Union[tuple(query.codec for query in queries)],
            },
        )

    def __call__(self, event: Union[Row, Table]) -> None:
        conn = self._connection_factory()
        self._run_query(conn, event)

    def _run_query(self, conn: Connection, event: Union[Row, Table]) -> None:
        query = self._queries[type(event)]
        if isinstance(event, Row):
            table = Table(rows=[event])
        elif isinstance(event, Table):
            table = event
        else:
            raise ValueError('Wrong event type {0}'.format(event))
        _exec_queries(conn, query, table)

    def close(self) -> None:
        self._closed = True
